"""Translate control flows to linear flows with goto-statements and co-routines."""
import abc
import typing
from typing import (
    Optional,
    Union,
    Sequence,
    List,
    Tuple,
    Set,
    Dict,
    cast,
    overload,
)

from icontract import require, ensure

from aas_core_codegen import common
from aas_core_codegen.common import (
    Stripped,
    indent_but_first_line,
    pairwise,
    iterate_except_first,
)
from aas_core_codegen.yielding import flow as yielding_flow


class Statement(abc.ABC):
    """Represent an abstract statement in a co-routine."""

    @require(lambda label: not (label is not None) or label >= 0)
    def __init__(self, label: Optional[int] = None) -> None:
        """Initialize with the given values."""
        self.label = label


class Command(Statement):
    """Represent a command which contains no jumps."""

    def __init__(self, code: Stripped, label: Optional[int] = None) -> None:
        """Initialize with the given values."""
        Statement.__init__(self, label=label)

        self.code = code


class If(Statement):
    """Represent a conditional jump, where we jump based on the condition."""

    def __init__(
        self,
        condition: Stripped,
        on_true: Optional[int] = None,
        on_false: Optional[int] = None,
        label: Optional[int] = None,
    ) -> None:
        """Initialize with the given values."""
        Statement.__init__(self, label=label)

        self.condition = condition
        self.on_true = on_true
        self.on_false = on_false


class Jump(Statement):
    """Represent an unconditional jump in the execution."""

    def __init__(self, target: int, label: Optional[int] = None) -> None:
        """Initialize with the given values."""
        Statement.__init__(self, label=label)

        self.target = target


class Yield(Statement):
    """Represent a yield statement, where the next co-routine should take over."""

    def __init__(self, label: Optional[int] = None) -> None:
        """Initialize with the given values."""
        Statement.__init__(self, label=label)


class Noop(Statement):
    """Represent a no-operation statement, which does nothing."""

    def __init__(
        self, label: Optional[int] = None, comment: Optional[Stripped] = None
    ) -> None:
        """Initialize with the given values."""
        Statement.__init__(self, label=label)
        self.comment = comment


StatementUnion = Union[Command, If, Jump, Yield, Noop]
common.assert_union_of_descendants_exhaustive(
    union=StatementUnion, base_class=Statement
)


def _dump_command_without_label(command: Command) -> str:
    return command.code


def _dump_if_without_label(if_statement: If) -> str:
    indent = "  "
    blocks = []  # type: List[str]
    if "\n" in if_statement.condition:
        blocks.append(
            f"""\
if
{indent}{indent_but_first_line(if_statement.condition, indent)}"""
        )
    else:
        blocks.append(f"if {if_statement.condition}")

    if if_statement.on_true is not None:
        blocks.append(f"is true, jump to {if_statement.on_true}")

    if if_statement.on_false is not None:
        blocks.append(f"is false, jump to {if_statement.on_false}")

    return "\n".join(blocks)


def _dump_jump_without_label(jump: Jump) -> str:
    return f"jump {jump.target}"


def _dump_yield_without_label(yield_statement: Yield) -> str:
    return "yield"


def _dump_noop_without_label(noop_statement: Noop) -> str:
    if noop_statement.comment is not None:
        return f"noop - {noop_statement.comment}"

    return "noop"


_DUMP_WITHOUT_LABEL_DISPATCH = {
    Command: _dump_command_without_label,
    If: _dump_if_without_label,
    Jump: _dump_jump_without_label,
    Yield: _dump_yield_without_label,
    Noop: _dump_noop_without_label,
}
assert all(
    cls in _DUMP_WITHOUT_LABEL_DISPATCH for cls in typing.get_args(StatementUnion)
), "All statements covered in _DUMP_WITHOUT_LABEL_DISPATCH"


def _dump_without_label(statement: StatementUnion) -> str:
    _dump_without_label_func = _DUMP_WITHOUT_LABEL_DISPATCH[statement.__class__]
    result = _dump_without_label_func(statement)  # type: ignore
    assert isinstance(result, str)  # necessary for mypy
    return result


def dump(statements: Sequence[StatementUnion]) -> str:
    """Render the statements to a textual sequence of labeled code."""
    if len(statements) == 0:
        return ""

    max_label = None
    for statement in statements:
        if statement.label is not None:
            if max_label is None:
                max_label = statement.label
            else:
                max_label = statement.label

    if max_label is None:
        label_characters = 0
    else:
        label_characters = len(str(max_label))

    label_format = f"{{:{label_characters}d}}: "
    empty_label = " " * label_characters + ": "
    indent = " " * (label_characters + 2)

    blocks = []  # type: List[str]
    for statement in statements:
        if statement.label is not None:
            label_str = label_format.format(statement.label)
        else:
            label_str = empty_label

        code = _dump_without_label(statement)

        blocks.append(
            f"""\
{label_str}{indent_but_first_line(code, indent)}"""
        )

    return "\n".join(blocks)


def _linearize_command(
    command: yielding_flow.Command, label: int
) -> Tuple[List[StatementUnion], int]:
    """
    Linearize the command with the given label.

    Return the linearization and the next available label.
    """
    return [Command(code=command.code, label=label)], label + 1


def _linearize_if_true(
    if_true_node: yielding_flow.IfTrue, label: int
) -> Tuple[List[StatementUnion], int]:
    """
    Linearize recursively the if-node in the control flow with the given label.

    Return the linearization and the next available label.
    """
    # NOTE (mristin, 2023-10-20):
    # We will fix this if-statement at the end, once we know the labels for ``on_true``
    # and ``on_false``.
    if_statement = If(condition=if_true_node.condition, label=label)

    result = [if_statement]  # type: List[StatementUnion]
    label += 1

    if if_true_node.or_else is not None:
        body, label = _linearize_sequence(if_true_node.body, label)
        result.extend(body)

        # NOTE (mristin, 2023-10-20):
        # This jump will be fixed after we know the exact target.
        jump_to_done_after_body = Jump(target=-1, label=label)
        result.append(jump_to_done_after_body)
        label += 1

        on_false = label
        or_else, label = _linearize_sequence(if_true_node.or_else, label)
        result.extend(or_else)

        done_label = label
        result.append(Noop(label=label))
        label += 1

        jump_to_done_after_body.target = done_label

        if_statement.on_false = on_false

    else:
        body, label = _linearize_sequence(if_true_node.body, label)
        if len(body) == 0:
            result.append(Noop(label=label))
        result.extend(body)

        done_label = label
        result.append(Noop(label=label))
        label += 1

        if_statement.on_false = done_label

    return result, label


def _linearize_if_false(
    if_false_node: yielding_flow.IfFalse, label: int
) -> Tuple[List[StatementUnion], int]:
    """
    Linearize recursively the if-node in the control flow with the given label.

    Return the linearization and the next available label.
    """
    # NOTE (mristin, 2023-10-20):
    # We will fix this if-statement at the end, once we know the labels for ``on_true``
    # and ``on_false``.
    if_statement = If(condition=if_false_node.condition, label=label)

    result = [if_statement]  # type: List[StatementUnion]
    label += 1

    if if_false_node.or_else is not None:
        body, label = _linearize_sequence(if_false_node.body, label)
        result.extend(body)

        # NOTE (mristin, 2023-10-22):
        # This jump will be fixed after we know the exact target.
        jump_to_done_after_body = Jump(target=-1, label=label)
        result.append(jump_to_done_after_body)
        label += 1

        on_true = label
        or_else, label = _linearize_sequence(if_false_node.or_else, label)
        result.extend(or_else)

        done_label = label
        result.append(Noop(label=label))
        label += 1

        jump_to_done_after_body.target = done_label

        if_statement.on_true = on_true

    else:
        body, label = _linearize_sequence(if_false_node.body, label)
        if len(body) == 0:
            result.append(Noop(label=label))
        result.extend(body)

        done_label = label
        result.append(Noop(label=label))
        label += 1

        if_statement.on_true = done_label

    return result, label


def _linearize_for(
    for_node: yielding_flow.For, label: int
) -> Tuple[List[StatementUnion], int]:
    """
    Linearize recursively the for-node in the control flow with the given label.

    Return the linearization and the next available label.
    """
    result = []  # type: List[StatementUnion]

    if for_node.init is not None:
        result.append(Command(for_node.init, label=label))
        label += 1

    # NOTE (mristin, 2023-10-20):
    # We will fix this if-statement at the end, once we know the label for ``on_true``.
    if_statement = If(condition=for_node.condition, label=label)

    result.append(if_statement)
    label += 1

    body, label = _linearize_sequence(for_node.body, label=label)
    result.extend(body)

    result.append(Command(for_node.iteration, label=label))
    label += 1

    assert if_statement.label is not None, (
        "The label of the condition If-statement must have been set; otherwise "
        "we do not know where to jump at the end of a loop iteration."
    )
    result.append(Jump(target=if_statement.label, label=label))
    label += 1

    done_label = label
    result.append(Noop(label=done_label))
    label += 1

    if_statement.on_false = done_label

    return result, label


def _linearize_while(
    while_node: yielding_flow.While, label: int
) -> Tuple[List[StatementUnion], int]:
    """
    Linearize the while-node in the control flow with the given label.

    Return the linearization and the next available label.
    """
    result = []  # type: List[StatementUnion]

    # NOTE (mristin, 2023-10-28):
    # We will fix this if-statement at the end, once we know the label for ``on_true``.
    if_statement = If(condition=while_node.condition, label=label)
    label += 1
    result.append(if_statement)

    body, label = _linearize_sequence(while_node.body, label=label)
    result.extend(body)

    assert if_statement.label is not None, (
        "The label of the condition If-statement must have been set; otherwise "
        "we do not know where to jump at the end of a loop iteration."
    )

    result.append(Jump(target=if_statement.label, label=label))
    label += 1

    done_label = label
    result.append(Noop(label=done_label))
    label += 1

    if_statement.on_false = done_label

    return result, label


def _linearize_yield(
    yield_node: yielding_flow.Yield, label: int
) -> Tuple[List[StatementUnion], int]:
    """
    Linearize the yield node in the control flow with the given label.

    Return the linearization and the next available label.
    """
    return [Yield(label=label)], label + 1


_LINEARIZE_DISPATCH = {
    yielding_flow.Command: _linearize_command,
    yielding_flow.IfTrue: _linearize_if_true,
    yielding_flow.IfFalse: _linearize_if_false,
    yielding_flow.For: _linearize_for,
    yielding_flow.While: _linearize_while,
    yielding_flow.Yield: _linearize_yield,
}
assert all(
    cls in _LINEARIZE_DISPATCH for cls in typing.get_args(yielding_flow.Node)
), "All classes covered in _LINEARIZE_DISPATCH"


def _linearize_node(
    node: yielding_flow.Node, label: int
) -> Tuple[List[StatementUnion], int]:
    """
    Linearize the yield node in the control flow with the given label.

    Return the linearization and the next available label.
    """
    linearize_func = _LINEARIZE_DISPATCH[node.__class__]
    result = linearize_func(node, label)  # type: ignore

    assert isinstance(result, tuple)
    assert isinstance(result[0], list)
    assert isinstance(result[1], int)

    return linearize_func(node, label)  # type: ignore


@ensure(
    lambda sequence, label, result: not (len(sequence) == 0)
    or (len(result[0]) == 0 and label == result[1]),
    "Empty sequence implies no statements and no change in labels",
)
def _linearize_sequence(
    sequence: Sequence[yielding_flow.Node], label: int
) -> Tuple[List[StatementUnion], int]:
    """
    Linearize recursively the control flow sequence, starting with the given label.
    :param sequence:
    :return:
    """
    result = []  # type: List[StatementUnion]
    for node in sequence:
        statements, label = _linearize_node(node, label)
        result.extend(statements)

    return result, label


def _linearize_control_flow(flow: Sequence[yielding_flow.Node]) -> List[StatementUnion]:
    """Compile the control flow into a linearized sequence of statements."""
    statements, _ = _linearize_sequence(flow, label=0)
    return statements


def _collect_targets(statements: Sequence[StatementUnion]) -> Set[int]:
    """Collect all the targets over all the sequences."""
    result = set()  # type: Set[int]
    for statement in statements:
        if isinstance(statement, Jump):
            result.add(statement.target)

        elif isinstance(statement, If):
            if statement.on_true is not None:
                result.add(statement.on_true)

            if statement.on_false is not None:
                result.add(statement.on_false)
        else:
            pass

    return result


def _remove_redundant_labels_in_place(statements: Sequence[StatementUnion]) -> None:
    """Remove labels from statements which are never a target."""
    target_set = _collect_targets(statements)
    for statement in statements:
        if statement.label not in target_set:
            statement.label = None


def _remove_noops_in_place(statements: List[StatementUnion]) -> List[StatementUnion]:
    """
    Remove no-ops and re-wire the targets in-place.

    The input statements list is invalidated, and should not be used after the call
    to this function.
    """
    # NOTE (mristin, 2023-10-20):
    # We can safely remove all no-ops which do not have a label, since they are
    # not targeted at all.
    statements = [
        statement
        for statement in statements
        if not isinstance(statement, Noop) or statement.label is not None
    ]

    # NOTE (mristin, 2023-10-20):
    # We will iterate through the statements now, map all the targets to the new
    # labels, and mark the no-ops for removal by unsetting their labels.

    old_to_new_target = dict()  # type: Dict[int, int]
    noop_block = []  # type: List[Noop]

    for statement in statements:
        if isinstance(statement, Noop):
            noop_block.append(statement)
        else:
            if len(noop_block) == 0:
                continue

            # NOTE (mristin, 2023-10-20):
            # If the statement does not have a label, we arbitrarily assign the label
            # of the first no-op in the block.
            if statement.label is None:
                assert noop_block[0].label is not None, (
                    "The label of the first no-op statement is always expected "
                    "as it starts a block."
                )

                statement.label = noop_block[0].label

            for noop in noop_block:
                assert noop.label is not None, (
                    "We have removed non-labeled no-ops before, "
                    "so all remaining no-ops must have a label."
                )

                assert (
                    statement.label is not None
                ), "The statement must have been set before so that we can map it."

                old_to_new_target[noop.label] = statement.label
                noop.label = None

            noop_block = []

    if len(noop_block) > 1:
        # NOTE (mristin, 2023-10-20):
        # This is a trailing no-op block. We simply reduce it to one no-op.
        iter_noop_block = iter(noop_block)
        next(iter_noop_block)

        for noop in iter_noop_block:
            assert noop.label is not None, (
                "We have removed non-labeled no-ops before, "
                "so all remaining no-ops must have a label."
            )

            assert noop_block[0].label is not None, (
                "Since all no-op statements must have a label, "
                "the first no-op in the block must have a label as well."
            )

            old_to_new_target[noop.label] = noop_block[0].label
            noop.label = None

    # NOTE (mristin, 2023-10-20):
    # We marked all no-ops for removal by unsetting their label.
    statements = [
        statement
        for statement in statements
        if not isinstance(statement, Noop) or statement.label is not None
    ]

    # NOTE (mristin, 2023-10-20):
    # Now we have to re-wire the targets.
    for statement in statements:
        if isinstance(statement, If):
            if statement.on_true is not None and statement.on_true in old_to_new_target:
                statement.on_true = old_to_new_target[statement.on_true]

            if (
                statement.on_false is not None
                and statement.on_false in old_to_new_target
            ):
                statement.on_false = old_to_new_target[statement.on_false]

        elif isinstance(statement, Jump):
            if statement.target in old_to_new_target:
                statement.target = old_to_new_target[statement.target]

        else:
            pass

    return statements


def _compress_in_place(statements: List[StatementUnion]) -> List[StatementUnion]:
    """
    Remove redundant statements and labels.

    The input statements list is invalidated, and should not be used after the call
    to this function.
    """
    _remove_redundant_labels_in_place(statements)
    statements = _remove_noops_in_place(statements)
    return statements


def _fix_labels_in_place(statements: List[StatementUnion]) -> None:
    """
    Go through statements and re-wire the labels so that they are consecutive.

    Moreover, we make sure that there is a label after each yield and at the first
    statement, so that we can split statements in blocks by labels.
    """
    if len(statements) == 0:
        return

    label = (
        max(
            statement.label if statement.label is not None else 0
            for statement in statements
        )
        + 1
    )

    # NOTE (mristin, 2023-10-20):
    # We simply set the first label to some arbitrary number and fix it later.
    if statements[0].label is None:
        statements[0].label = label
        label += 1

    # NOTE (mristin, 2023-10-21):
    # We add a label after each yield so that we can split the statements in block,
    # where each block starts with a statement label.
    for previous, current in pairwise(statements):
        if isinstance(previous, Yield) and current.label is None:
            current.label = label
            label += 1

    # NOTE (mristin, 2023-10-21):
    # Now reset all labels so that they are consecutive.

    label = 0

    old_to_new_label = dict()  # type: Dict[int, int]
    for statement in statements:
        if statement.label is not None:
            old_to_new_label[statement.label] = label
            label += 1

    for statement in statements:
        if statement.label is not None:
            statement.label = old_to_new_label[statement.label]

        if isinstance(statement, If):
            if statement.on_true is not None and statement.on_true in old_to_new_label:
                statement.on_true = old_to_new_label[statement.on_true]

            if (
                statement.on_false is not None
                and statement.on_false in old_to_new_label
            ):
                statement.on_false = old_to_new_label[statement.on_false]

        elif isinstance(statement, Jump):
            if statement.target in old_to_new_label:
                statement.target = old_to_new_label[statement.target]

        else:
            pass


class Subroutine(Sequence[StatementUnion]):
    """Capture a subroutine which can execute between the yields."""

    # fmt: off
    @require(
        lambda statements:
        statements[0].label is not None
        and all(
            statement.label is None
            for statement in iterate_except_first(statements)
        ),
        "Only the first statement should have the label defined, "
        "and the remainder of the statements should not."
    )
    @require(
        lambda statements: len(statements) > 0,
        "Empty subroutines are ill-defined"
    )
    # fmt: on
    def __new__(cls, statements: Sequence[StatementUnion]) -> "Subroutine":
        return cast(Subroutine, statements)

    @overload
    def __getitem__(self, index: int) -> StatementUnion:
        raise NotImplementedError("Only for type annotations")

    @overload
    def __getitem__(self, index: slice) -> "Subroutine":
        raise NotImplementedError("Only for type annotations")

    def __getitem__(
        self, index: Union[int, slice]
    ) -> Union[StatementUnion, "Subroutine"]:
        raise NotImplementedError("Only for type annotations")

    def __len__(self) -> int:
        raise NotImplementedError("Only for type annotations")


def _split_in_subroutines(statements: Sequence[StatementUnion]) -> List[Subroutine]:
    """Split the statements in blocks on each label."""
    result = []  # type: List[Subroutine]

    block = []  # type: List[StatementUnion]
    for statement in statements:
        if statement.label is not None:
            if len(block) > 0:
                result.append(Subroutine(block))

            block = [statement]

        else:
            block.append(statement)

    if len(block) > 0:
        result.append(Subroutine(block))

    return result


# fmt: off
@ensure(
    lambda result:
    all(
        subroutine[0].label + 1 == next_subroutine[0].label
        for subroutine, next_subroutine in pairwise(result)
    ),
    "Subroutine labels are consecutively increasing"
)
# fmt: on
def linearize_to_subroutines(flow: Sequence[yielding_flow.Node]) -> List[Subroutine]:
    """Linearize the control flow and split it in subroutines on each yield."""
    if len(flow) == 0:
        return []

    statements = _linearize_control_flow(flow)
    statements = _compress_in_place(statements)
    _fix_labels_in_place(statements)
    subroutines = _split_in_subroutines(statements)

    return subroutines



"""Generate instructions for a virtual machine for matching regular expressions."""

from aas_core_codegen.cpp.pattern import _generate

generate_header = _generate.generate_header
generate_implementation = _generate.generate_implementation



"""Generate Java code for copying in memory based on the intermediate representation."""

import io
import textwrap
from typing import Tuple, Optional, List

from icontract import ensure

from aas_core_codegen import intermediate, specific_implementations
from aas_core_codegen.common import (
    Error,
    Stripped,
    Identifier,
    indent_but_first_line,
    assert_never,
)
from aas_core_codegen.java import (
    common as java_common,
    naming as java_naming,
)
from aas_core_codegen.csharp.common import (
    INDENT as I,
    INDENT2 as II,
)


# region Generate


def _generate_shallow_copy_transform_method(
    cls: intermediate.ConcreteClass,
) -> Stripped:
    """Generate the method in the transformer to make a shallow copy of ``cls''."""
    property_names = [prop.name for prop in cls.properties]
    constructor_argument_names = [arg.name for arg in cls.constructor.arguments]

    # fmt: off
    assert (
        set(prop.name for prop in cls.properties)
        == set(arg.name for arg in cls.constructor.arguments)
    ), (
        f"Expected the properties to coincide with constructor arguments, "
        f"but they do not for {cls.name!r}:"
        f"{property_names=}, {constructor_argument_names=}"
    )
    # fmt: on

    cls_name = java_naming.class_name(cls.name)

    if len(cls.constructor.arguments) == 0:
        return_statement = Stripped(f"return new {cls_name}();")
    else:
        constructor_arg_exprs = []  # type: List[str]
        for arg in cls.constructor.arguments:
            if isinstance(arg.type_annotation, intermediate.OptionalTypeAnnotation):
                getter_name = java_naming.getter_name(arg.name)
                constructor_arg_exprs.append(f"that.{getter_name}().orElse(null)")
            else:
                getter_name = java_naming.getter_name(arg.name)
                constructor_arg_exprs.append(f"that.{getter_name}()")

        args_joined = ", ".join(constructor_arg_exprs)
        return_statement = Stripped(f"return new {cls_name}({args_joined});")

        if len(return_statement) > 70:
            args_joined = ",\n".join(constructor_arg_exprs)
            return_statement = Stripped(
                f"""\
return new {cls_name}(
{I}{indent_but_first_line(args_joined, I)});"""
            )

    interface_name = java_naming.interface_name(cls.name)
    transform_name = java_naming.method_name(Identifier(f"transform_{cls.name}"))

    return Stripped(
        f"""\
@Override
public IClass {transform_name}(
{I}{interface_name} that
) {{
{I}{indent_but_first_line(return_statement, I)}
}}"""
    )


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_shallow_copier(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the transformer which makes shallow copies."""
    errors = []  # type: List[Error]

    blocks = []  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        if cls.is_implementation_specific:
            implementation_key = specific_implementations.ImplementationKey(
                f"Copying/ShallowCopier/transform_{cls.name}.java"
            )

            implementation = spec_impls.get(implementation_key, None)
            if implementation is None:
                errors.append(
                    Error(
                        cls.parsed.node,
                        f"The snippet for making shallow copies is missing "
                        f"for the implementation-specific "
                        f"class {cls.name}: {implementation_key}",
                    )
                )
                continue

            blocks.append(spec_impls[implementation_key])
        else:
            blocks.append(_generate_shallow_copy_transform_method(cls=cls))

    writer = io.StringIO()
    writer.write(
        """\
/**
 * Dispatch the making of shallow copies.
 */
private static class ShallowCopier extends AbstractTransformer<IClass> {
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue()), None


def _generate_deep_copy_transform_method(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate the method in the transformer to make a deep copy of ``cls''."""
    property_names = [prop.name for prop in cls.properties]
    constructor_argument_names = [arg.name for arg in cls.constructor.arguments]

    # fmt: off
    assert (
            set(prop.name for prop in cls.properties)
            == set(arg.name for arg in cls.constructor.arguments)
    ), (
        f"Expected the properties to coincide with constructor arguments, "
        f"but they do not for {cls.name!r}:"
        f"{property_names=}, {constructor_argument_names=}"
    )
    # fmt: on

    cls_name = java_naming.class_name(cls.name)

    body_blocks = []  # type: List[Stripped]

    if len(cls.constructor.arguments) == 0:
        body_blocks.append(Stripped(f"return new {cls_name}();"))
    else:
        for arg in cls.constructor.arguments:
            getter_name = java_naming.getter_name(arg.name)
            optional = isinstance(
                arg.type_annotation, intermediate.OptionalTypeAnnotation
            )

            type_anno = intermediate.beneath_optional(arg.type_annotation)

            if not isinstance(type_anno, intermediate.ListTypeAnnotation):
                continue

            assert isinstance(
                type_anno.items, intermediate.OurTypeAnnotation
            ) and isinstance(type_anno.items.our_type, intermediate.Class), (
                "We handle only lists of classes in the deep copies at the "
                "moment. The meta-model does not contain any other lists, so "
                "we wanted to keep the code as simple as possible, and avoid "
                "unrolling. Please contact the developers if you need this "
                "feature."
            )

            # We need to prefix to avoid any possible naming conflicts.
            variable_name = java_naming.variable_name(Identifier(f"the_{arg.name}"))

            variable_type = java_common.generate_type(type_anno)

            inner_type = java_naming.interface_name(type_anno.items.our_type.name)

            if not optional:
                body_blocks.append(
                    Stripped(
                        f"""\
{variable_type} {variable_name} = new ArrayList<>(
{I}that.{getter_name}().size());
for ({inner_type} item : that.{getter_name}()) {{
{I}{variable_name}.add(deep(item));
}}"""
                    )
                )
            else:
                other_property_name = java_naming.variable_name(
                    Identifier(f"that_{arg.name}")
                )

                body_blocks.append(
                    Stripped(
                        f"""\
{variable_type} {other_property_name} =
{I}that.{getter_name}().orElse(null);
{variable_type} {variable_name} = null;
if ({other_property_name} != null) {{
{I}{variable_name} = new ArrayList<>(
{II}{other_property_name}.size());
{I}for ({inner_type} item : {other_property_name})
{I}{{
{II}{variable_name}.add(deep(item));
{I}}}
}}"""
                    )
                )

        constructor_arg_exprs = []  # type: List[str]
        for arg in cls.constructor.arguments:
            getter_name = java_naming.getter_name(arg.name)

            type_anno = intermediate.beneath_optional(arg.type_annotation)

            optional = isinstance(
                arg.type_annotation, intermediate.OptionalTypeAnnotation
            )

            if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
                if optional:
                    constructor_arg_exprs.append(
                        f"""that.{getter_name}().orElse(null)"""
                    )
                else:
                    constructor_arg_exprs.append(f"""that.{getter_name}()""")

            elif isinstance(type_anno, intermediate.OurTypeAnnotation):
                if isinstance(type_anno.our_type, intermediate.Enumeration):
                    if optional:
                        constructor_arg_exprs.append(
                            f"""that.{getter_name}().orElse(null)"""
                        )
                    else:
                        constructor_arg_exprs.append(f"""that.{getter_name}()""")

                elif isinstance(type_anno.our_type, intermediate.ConstrainedPrimitive):
                    if optional:
                        constructor_arg_exprs.append(
                            f"""that.{getter_name}().orElse(null)"""
                        )
                    else:
                        constructor_arg_exprs.append(f"""that.{getter_name}()""")

                elif isinstance(
                    type_anno.our_type,
                    (intermediate.AbstractClass, intermediate.ConcreteClass),
                ):
                    if optional:
                        constructor_arg_exprs.append(
                            f"""that.{getter_name}().orElse(null)"""
                        )
                    else:
                        constructor_arg_exprs.append(f"deep(that.{getter_name}())")
                else:
                    assert_never(type_anno.our_type)

            elif isinstance(type_anno, intermediate.ListTypeAnnotation):
                # See how this variable is computed above in the generated code.
                constructor_arg_exprs.append(
                    java_naming.variable_name(Identifier(f"the_{arg.name}"))
                )
            else:
                assert_never(type_anno)

        return_statement_writer = io.StringIO()

        return_statement_writer.write(f"return new {cls_name}(\n")

        for i, arg_expr in enumerate(constructor_arg_exprs):
            return_statement_writer.write(textwrap.indent(arg_expr, I))

            if i < len(constructor_arg_exprs) - 1:
                return_statement_writer.write(",\n")
            else:
                return_statement_writer.write("\n")

        return_statement_writer.write(");")

        body_blocks.append(Stripped(return_statement_writer.getvalue()))

    body_writer = io.StringIO()
    for i, body_block in enumerate(body_blocks):
        if i > 0:
            body_writer.write("\n\n")

        body_writer.write(body_block)

    interface_name = java_naming.interface_name(cls.name)
    transform_name = java_naming.method_name(Identifier(f"transform_{cls.name}"))

    return Stripped(
        f"""\
@Override
public IClass {transform_name} (
{I}{interface_name} that
) {{
{I}{indent_but_first_line(body_writer.getvalue(), I)}
}}"""
    )


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_deep_copier(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the transformer which makes deep copies."""
    errors = []  # type: List[Error]

    blocks = []  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        if cls.is_implementation_specific:
            implementation_key = specific_implementations.ImplementationKey(
                f"Copying/DeepCopier/transform_{cls.name}.java"
            )

            implementation = spec_impls.get(implementation_key, None)
            if implementation is None:
                errors.append(
                    Error(
                        cls.parsed.node,
                        f"This snippet for making deep copies is missing "
                        f"for the implementation-specific "
                        f"class {cls.name}: {implementation_key}",
                    )
                )
                continue

            blocks.append(spec_impls[implementation_key])
        else:
            blocks.append(_generate_deep_copy_transform_method(cls=cls))

    writer = io.StringIO()
    writer.write(
        """\
/** Dispatch the making of deep copies. */
private static class DeepCopier extends AbstractTransformer<IClass> {
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue()), None


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable,
    package: java_common.PackageIdentifier,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """
    Generate the Java code for copying instances in memory.

    The ``namespace`` defines the AAS Java namespace.
    """
    errors = []  # type: List[Error]

    imports = [
        Stripped("import java.util.List;"),
        Stripped("import java.util.ArrayList;"),
        Stripped(f"import {package}.types.{java_common.INTERFACE_PKG}.IClass;"),
        Stripped(f"import {package}.visitation.AbstractTransformer;"),
        Stripped(f"import {package}.types.impl.*;"),
        Stripped(f"import {package}.types.model.*;"),
    ]  # type: List[Stripped]

    # NOTE (empwilli, 2023-12-14):
    # We wrap the shallow and deep copying in generic methods to allow for easier
    # enforcement of runtime type safety for the client. Otherwise, if we directly
    # provided the transformer, the client would always need to make the casts, which
    # is cumbersome.

    copy_blocks = [
        Stripped(
            """private static final ShallowCopier shallowCopierInstance = new ShallowCopier();"""
        ),
        Stripped(
            """private static final DeepCopier deepCopierInstance = new DeepCopier();"""
        ),
        Stripped(
            f"""\
/**
 * Make a shallow copy of {{@code that}}.
 *
 * <p>All the properties are copied by reference. This includes also the lists.
 * Hence, a list property is copied by reference, and not, as sometimes might be
 * expected, as a new list of underlying references.
 *
 * @param that to be copied in a shallow manner
 */
@SuppressWarnings("unchecked")
public static <T extends IClass> T shallow(T that) {{
{I}return (T) shallowCopierInstance.transform(that);
}}"""
        ),
        Stripped(
            f"""\
/**
 * Make a recursively a deep copy of {{@code that}}.
 *
 * @param that to be deeply copied in a recursive manner
 */
@SuppressWarnings("unchecked")
public static <T extends IClass> T deep(T that) {{
{I}return (T) deepCopierInstance.transform(that);
}}"""
        ),
    ]  # type: List[Stripped]

    shallow_copier_block, shallow_errors = _generate_shallow_copier(
        symbol_table=symbol_table, spec_impls=spec_impls
    )
    if shallow_errors is not None:
        errors.extend(shallow_errors)
    else:
        assert shallow_copier_block is not None
        copy_blocks.append(shallow_copier_block)

    deep_copier_block, deep_errors = _generate_deep_copier(
        symbol_table=symbol_table, spec_impls=spec_impls
    )
    if deep_errors is not None:
        errors.extend(deep_errors)
    else:
        assert deep_copier_block is not None
        copy_blocks.append(deep_copier_block)

    if len(errors) > 0:
        return None, errors

    copying_writer = io.StringIO()
    copying_writer.write(
        """\
/**
 * Allow for making shallow and deep copies of AAS model instances.
 */
public class Copying
{
"""
    )

    for i, copy_block in enumerate(copy_blocks):
        if i > 0:
            copying_writer.write("\n\n")

        copying_writer.write(textwrap.indent(copy_block, I))

    copying_writer.write("\n}")

    blocks = [
        java_common.WARNING,
        Stripped(f"package {package}.copying;"),
        Stripped("\n".join(imports)),
        Stripped(f"{copying_writer.getvalue()}"),
        java_common.WARNING,
    ]  # type: List[Stripped]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        assert not block.startswith("\n")
        assert not block.endswith("\n")
        writer.write(block)

    writer.write("\n")

    return writer.getvalue(), None


# endregion



"""Generate the Java constants corresponding to the constants of the meta-model."""
import io
import textwrap
from typing import (
    List,
    Optional,
    Set,
    Tuple,
)

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    assert_never,
    Error,
    Stripped,
)
from aas_core_codegen.java import (
    common as java_common,
    naming as java_naming,
)
from aas_core_codegen.csharp.common import (
    INDENT as I,
    INDENT2 as II,
)

# region Generation


def _byte_array_as_expr(value: bytearray) -> Stripped:
    """Return a multi-line java code representing the ``value''."""
    writer = io.StringIO()

    writer.write("new byte[] {")

    for i, byte in enumerate(value):
        if i > 0:
            if i % 8 == 0:
                writer.write(",\n")
            else:
                writer.write(", ")

        writer.write(f"(byte) 0x{byte:02x}")

    writer.write("\n}")

    return Stripped(writer.getvalue())


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_constant_primitive(
    constant: intermediate.ConstantPrimitive,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the definition of a constant primitive."""
    constant_name = java_naming.property_name(constant.name)

    assert constant.a_type in java_common.PRIMITIVE_TYPE_MAP

    java_type = java_common.PRIMITIVE_TYPE_MAP[constant.a_type]

    literal: Stripped

    if constant.a_type is intermediate.PrimitiveType.BOOL:
        literal = Stripped("true") if constant.value else Stripped("false")

    elif constant.a_type is intermediate.PrimitiveType.INT:
        literal = Stripped(str(constant.value))

    elif constant.a_type is intermediate.PrimitiveType.FLOAT:
        # NOTE (empwilli, 2023-12-14):
        # We assume that the float constants are not meant to be all to precise.
        # Therefore, we use a string representation here. However, beware that we
        # might have to use a more precise representation in the future if the spec
        # change.
        literal = Stripped(str(constant.value))

    elif constant.a_type is intermediate.PrimitiveType.STR:
        assert isinstance(constant.value, str)
        literal = java_common.string_literal(constant.value)

    elif constant.a_type is intermediate.PrimitiveType.BYTEARRAY:
        assert isinstance(constant.value, bytearray)
        literal = _byte_array_as_expr(value=constant.value)

    else:
        assert_never(constant.a_type)

    return (
        Stripped(f"public static const {java_type} {constant_name} = {literal};"),
        None,
    )


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_constant_set_of_primitives(
    constant: intermediate.ConstantSetOfPrimitives,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the definition of a constant set of primitives."""
    constant_name = java_naming.property_name(constant.name)

    writer = io.StringIO()

    java_type = java_common.PRIMITIVE_TYPE_MAP[constant.a_type]

    writer.write(
        f"""\
public static final Set<{java_type}> {constant_name} = Stream.of(
"""
    )

    if constant.a_type is intermediate.PrimitiveType.BOOL:
        for i, literal in enumerate(constant.literals):
            writer.write(textwrap.indent("true" if literal.value else "false", II))

            if i < len(constant.literals) - 1:
                writer.write(",\n")
            else:
                writer.write("\n")

    elif constant.a_type is intermediate.PrimitiveType.INT:
        for i, literal in enumerate(constant.literals):
            writer.write(textwrap.indent(str(literal.value), II))

            if i < len(constant.literals) - 1:
                writer.write(",\n")
            else:
                writer.write("\n")

    elif constant.a_type is intermediate.PrimitiveType.FLOAT:
        # NOTE (empwilli, 2023-12-14):
        # We assume that the float constants are not meant to be all to precise.
        # Therefore, we use a string representation here. However, beware that we
        # might have to use a more precise representation in the future if the spec
        # change.

        for i, literal in enumerate(constant.literals):
            writer.write(textwrap.indent(str(literal.value), II))

            if i < len(constant.literals) - 1:
                writer.write(",\n")
            else:
                writer.write("\n")

    elif constant.a_type is intermediate.PrimitiveType.STR:
        for i, literal in enumerate(constant.literals):
            assert isinstance(literal.value, str)

            writer.write(textwrap.indent(java_common.string_literal(literal.value), II))

            if i < len(constant.literals) - 1:
                writer.write(",\n")
            else:
                writer.write("\n")

    elif constant.a_type is intermediate.PrimitiveType.BYTEARRAY:
        for i, literal in enumerate(constant.literals):
            assert isinstance(literal.value, bytearray)

            writer.write(textwrap.indent(_byte_array_as_expr(literal.value), II))

            if i < len(constant.literals) - 1:
                writer.write(",\n")
            else:
                writer.write("\n")

    else:
        assert_never(constant.a_type)

    writer.write(""").collect(ImmutableCollector.toImmutableSet());""")

    return Stripped(writer.getvalue()), None


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_constant_set_of_enumeration_literals(
    constant: intermediate.ConstantSetOfEnumerationLiterals,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the definition of a constant set of enumeration literals."""
    constant_name = java_naming.property_name(constant.name)
    enum_name = java_naming.enum_name(constant.enumeration.name)

    writer = io.StringIO()

    # NOTE (empwilli, 2023-12-14):
    # We make the sets of enumeration literals work on nullables to avoid checking
    # nullability all the time in the code. This gives a bit less performant code,
    # but a much more readable one.

    writer.write(
        f"""\
public static final Set<{enum_name}> {constant_name} = Stream.of(
"""
    )

    for i, literal in enumerate(constant.literals):
        literal_name = java_naming.enum_literal_name(literal.name)

        writer.write(textwrap.indent(f"{enum_name}.{literal_name}", II))

        if i < len(constant.literals) - 1:
            writer.write(",\n")
        else:
            writer.write("\n")

    writer.write(""").collect(ImmutableCollector.toImmutableSet());""")

    return Stripped(writer.getvalue()), None


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable,
    package: java_common.PackageIdentifier,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """
    Generate the Java code of the constants based on the symbol table.

    The ``package`` defines the root Java package.
    """
    enum_imports = set()  # type: Set[Stripped]

    for constant in symbol_table.constants:
        if not isinstance(constant, intermediate.ConstantSetOfEnumerationLiterals):
            continue

        enum_name = java_naming.enum_name(constant.enumeration.name)

        enum_imports.add(
            Stripped(f"import {package}.types.{java_common.ENUM_PKG}.{enum_name};")
        )

    enum_imports_block = Stripped("\n".join(sorted(enum_imports)))

    constants_blocks = []  # type: List[Stripped]

    errors = []  # type: List[Error]

    for constant in symbol_table.constants:
        constants_block: Optional[Stripped]
        error: Optional[Error]

        if isinstance(constant, intermediate.ConstantPrimitive):
            constants_block, error = _generate_constant_primitive(constant)
        elif isinstance(constant, intermediate.ConstantSetOfPrimitives):
            constants_block, error = _generate_constant_set_of_primitives(constant)
        elif isinstance(constant, intermediate.ConstantSetOfEnumerationLiterals):
            constants_block, error = _generate_constant_set_of_enumeration_literals(
                constant
            )
        else:
            assert_never(constant)

        if error is not None:
            errors.append(error)
            continue

        assert constants_block is not None
        constants_blocks.append(constants_block)

    if len(errors) > 0:
        return None, errors

    constants_writer = io.StringIO()
    constants_writer.write(
        """\
/**
 * Provide constant values of the meta-model.
 */
public class Constants {
"""
    )

    for i, constants_block in enumerate(constants_blocks):
        assert len(constants_block) > 0

        if i > 0:
            constants_writer.write("\n\n")

        constants_writer.write(textwrap.indent(constants_block, I))

    constants_writer.write("\n}")

    blocks = [
        java_common.WARNING,
        Stripped(
            f"""\
package {package}.constants;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.stream.Collector;
import java.util.stream.Stream;"""
        ),
        enum_imports_block,
        Stripped(
            f"""\
// Helper to generate read-only collections with less boilerplate.
// See: https://stackoverflow.com/a/37406054
class ImmutableCollector {{
    public static <T> Collector<T, Set<T>, Set<T>> toImmutableSet() {{
        return Collector.of(HashSet::new, Set::add, (l, r) -> {{
            l.addAll(r);
            return l;
        }}, Collections::unmodifiableSet);
    }}
}}

{constants_writer.getvalue()}"""
        ),
        java_common.WARNING,
    ]  # type: List[Stripped]

    out = io.StringIO()
    for i, constants_block in enumerate(blocks):
        if i > 0:
            out.write("\n\n")

        out.write(constants_block)

    out.write("\n")

    return out.getvalue(), None


# endregion



"""Generate TypeScript code of common functions by including the code directly."""

import io

from icontract import ensure

from aas_core_codegen.common import (
    Stripped,
)
from aas_core_codegen.typescript import (
    common as typescript_common,
)
from aas_core_codegen.typescript.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
)


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate() -> str:
    """Generate the TypeScript code for common functions."""
    blocks = [
        Stripped(
            """\
/**
 * Provide common functions shared among the modules.
 */"""
        ),
        typescript_common.WARNING,
        Stripped(
            f"""\
/**
 * Create an iterator over the given range of numbers.
 *
 * @param start - inclusive start of the range
 * @param end - exclusive end of the range
 * @returns iterator over the range
 */
// eslint-disable-next-line @typescript-eslint/no-unused-vars
export function *range(start: number, end: number): IterableIterator<number> {{
{I}for (let i = start; i < end; i++) {{
{II}yield i;
{I}}}
}}"""
        ),
        Stripped(
            f"""\
/**
 * Retrieve the `index`-th item from the `array`.
 *
 * @remarks
 * This is a fill for `Array.prototype.at`.
 * See: https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array/at
 *
 * @param array - to get the element from
 * @param index - zero-based index of the `array`. Negative index counts back.
 * @returns item, or `undefined` if `index` out-of-bound
 * @typeParam T - type of the array items
 */
export function at<T>(
{I}array: Array<T>,
{I}index: number
) {{
{I}if (index < 0) {{
{II}return array[array.length + index];
{I}}} else {{
{II}return array[index];
{I}}}
}}"""
        ),
        Stripped(
            f"""\
/**
 * Check that all the values of the iterable are `true`.
 *
 * @param iterable - to iterate over
 * @returns `true` if all values in `iterable` are set
 */
export function every<T>(
{I}iterable: Iterable<T>
): boolean {{
{I}// NOTE (mristin, 2022-11-24):
{I}// We introduce this function so that we can keep the constraint verification
{I}// purely functional. Unfortunately, `every` and `some` are only available
{I}// in arrays and not in `IterableIterator`.

{I}for (const item of iterable) {{
{II}if (!item) {{
{III}return false;
{II}}}
{I}}}

{I}return true;
}}"""
        ),
        Stripped(
            f"""\
/**
 * Check that at least one value of the iterable is `true`.
 *
 * @param iterable - to iterate over
 * @returns `true` if at least one value in `iterable` is set
 */
export function some<T>(
{I}iterable: Iterable<T>
): boolean {{
{I}// NOTE (mristin, 2022-11-24):
{I}// We introduce this function so that we can keep the constraint verification
{I}// purely functional. Unfortunately, `every` and `some` are only available
{I}// in arrays and not in `IterableIterator`.

{I}for (const item of iterable) {{
{II}if (item) {{
{III}return true;
{II}}}
{I}}}

{I}return false;
}}"""
        ),
        Stripped(
            f"""\
/**
 * Map the items of an iterable.
 *
 * @param iterable - to be mapped
 * @param mappingFunction - to be applied on `iterable`
 * @returns mapped items of `iterable`
 * @typeParam S - type of an item of the `iterable`
 * @typeParam T - type of the transformed item of the `iterable`
 */
export function *map<S, T>(
{I}iterable: Iterable<S>,
{I}mappingFunction: (item: S) => T
): IterableIterator<T> {{
{I}// NOTE (mristin, 2022-11-24):
{I}// We introduce this function so that we can keep the constraint verification
{I}// purely functional.

{I}for (const item of iterable) {{
{II}yield mappingFunction(item);
{I}}}
}}"""
        ),
        Stripped(
            f"""\
/**
 * Represent either a result, or an error.
 *
 * @typeParam ValueT - type of the resulting value
 * @typeParam ErrorT - type of the error
 */
export class Either<ValueT, ErrorT> {{
{I}/**
{I} * value if something successful
{I} */
{I}readonly value: ValueT | null;

{I}/**
{I} * error if something failed
{I} */
{I}readonly error: ErrorT | null;

{I}/**
{I} * Assert that value is set and return it.
{I} *
{I} * @returns {{@link value}}, or throw if `null`
{I} */
{I}mustValue(): ValueT {{
{II}if (this.value === null) {{
{III}throw new Error("Expected value to be set, but it was null");
{II}}}
{II}return this.value;
{I}}}

{I}constructor(value: ValueT | null, error: ErrorT | null) {{
{II}if (value === null && error === null) {{
{III}throw new Error("Unexpected both value and error null in an Either");
{II}}}

{II}if (value !== null && error !== null) {{
{III}throw new Error("Unexpected both value and error non-null in an Either");
{II}}}

{II}this.value = value;
{II}this.error = error;
{I}}}
}}"""
        ),
        # pylint: disable=line-too-long
        Stripped(
            f"""\
const BASE64_CHARS = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
const BASE64_LOOKUP = new Uint8Array(256);

// NOTE (mristin, 2022-11-25):
// Initialize to 255 so that we can detect invalid values in the input during decoding.
for (let i = 0; i < BASE64_LOOKUP.length; i++) {{
{I}BASE64_LOOKUP[i] = 255;
}}

// NOTE (mristin, 2022-11-25):
// Initialize valid values to the corresponding decoding points.
for (let i = 0; i < BASE64_CHARS.length; i++) {{
{I}BASE64_LOOKUP[BASE64_CHARS.charCodeAt(i)] = i;
}}

/**
 * Encode a byte array in base64.
 *
 * @remarks
 * We provide our own implementation so that we do not run into compatibility
 * issues with node.js, different browsers etc.
 * See:
 * https://stackoverflow.com/questions/21797299/convert-base64-string-to-arraybuffer
 *
 * @param bytes - to be encoded
 * @returns `bytes` encoded as base64 text
 */
export function base64Encode(bytes: Uint8Array): string {{
{I}// NOTE (mristin, 2022-11-25):
{I}// This implementation is vaguely based on:
{I}// https://github.com/danguer/blog-examples/blob/master/js/base64-binary.js,
{I}// https://github.com/niklasvh/base64-arraybuffer/blob/master/src/index.ts and
{I}// https://github.com/beatgammit/base64-js/blob/master/index.js.

{I}// NOTE (mristin, 2022-11-25):
{I}// We assume that string concatenation is actually *faster* than joining an array
{I}// of strings, see:
{I}// https://stackoverflow.com/questions/51185/are-javascript-strings-immutable-do-i-need-a-string-builder-in-javascript

{I}if (bytes.length === 0) {{
{II}return "";
{I}}}

{I}let encoded = '';
{I}const len = bytes.length;

{I}for (let i = 0; i < len; i += 3) {{
{II}encoded += BASE64_CHARS[bytes[i] >> 2];
{II}encoded += BASE64_CHARS[((bytes[i] & 3) << 4) | (bytes[i + 1] >> 4)];
{II}encoded += BASE64_CHARS[((bytes[i + 1] & 15) << 2) | (bytes[i + 2] >> 6)];
{II}encoded += BASE64_CHARS[bytes[i + 2] & 63];
{I}}}

{II}// NOTE (mristin, 2022-11-25):
{II}// We assume here that `substring` will be optimized for cases where we do not keep
{II}// the original reference to the string. We tested a bit with
{II}// https://www.measurethat.net/.

{I}if (len % 3 === 2) {{
{II}encoded = encoded.substring(0, encoded.length - 1) + '=';
{I}}} else if (len % 3 === 1) {{
{II}encoded = encoded.substring(0, encoded.length - 2) + '==';
{I}}} else {{
{II}// No padding is necessary.
{I}}}

{I}return encoded;
}}"""
        ),
        # pylint: enable=line-too-long
        Stripped(
            f"""\
/**
 * Decode a base64-encoded byte array.
 *
 * @remarks
 * We provide our own implementation so that we do not run into compatibility
 * issues with node.js, different browsers etc.
 * See:
 * https://stackoverflow.com/questions/21797299/convert-base64-string-to-arraybuffer
 *
 * @param text - to be decoded
 * @returns either the array or an error, if `text` is not a valid base64 encoding
 */
export function base64Decode(text: string): Either<Uint8Array, string> {{
{I}// NOTE (mristin, 2022-11-25):
{I}// This implementation is vaguely based on:
{I}// https://github.com/danguer/blog-examples/blob/master/js/base64-binary.js,
{I}// https://github.com/niklasvh/base64-arraybuffer/blob/master/src/index.ts and
{I}// https://github.com/beatgammit/base64-js/blob/master/index.js.

{I}const len = text.length;
{I}let lenWoPad = len;

{I}// NOTE (mristin, 2022-11-25):
{I}// Some implementations forget the padding, so we try to be robust and check
{I}// for the padding manually.
{I}let bytesLength = text.length * 0.75;
{I}if (text[len - 1] === '=') {{
{II}bytesLength--;
{II}lenWoPad--;
{II}if (text[len - 2] === '=') {{
{III}bytesLength--;
{III}lenWoPad--;
{II}}}
{I}}}

{I}const bytes = new Uint8Array(bytesLength);

{I}const base64LookupLen = BASE64_LOOKUP.length;

{I}let pointer = 0;

{I}for (let i = 0; i < len; i += 4) {{
{II}// NOTE (mristin, 2022-11-25):
{II}// Admittedly, this is very verbose code, but we want to be efficient, so we
{II}// opted for performance over readability here.

{II}const charCode0 = text.charCodeAt(i);
{II}if (charCode0 >= base64LookupLen) {{
{III}return new Either<Uint8Array, string>(
{IIII}null,
{IIII}"Expected a valid character from base64-encoded string, " +
{IIIII}`but got at index ${{i}}: ${{text[i]}} (code: ${{charCode0}})`
{III});
{II}}}
{II}const encoded0 = BASE64_LOOKUP[charCode0];
{II}if (encoded0 === 255) {{
{III}return new Either<Uint8Array, string>(
{IIII}null,
{IIII}"Expected a valid character from base64-encoded string, " +
{IIIII}`but got at index ${{i}}: ${{text[i]}} (code: ${{charCode0}})`
{III});
{II}}}

{II}const charCode1 = text.charCodeAt(i + 1);
{II}if (charCode1 >= base64LookupLen) {{
{III}return new Either<Uint8Array, string>(
{IIII}null,
{IIII}"Expected a valid character from base64-encoded string, " +
{IIIII}`but got at index ${{i + 1}}: ${{text[i + 1]}} (code: ${{charCode1}})`
{III});
{II}}}
{II}const encoded1 = BASE64_LOOKUP[charCode1];
{II}if (encoded1 === 255) {{
{III}return new Either<Uint8Array, string>(
{IIII}null,
{IIII}"Expected a valid character from base64-encoded string, " +
{IIIII}`but got at index ${{i + 1}}: ${{text[i + 1]}} (code: ${{charCode1}})`
{III});
{II}}}

{II}// We map padding to 65, which is the value of "A".
{II}const charCode2 = i + 2 < lenWoPad ? text.charCodeAt(i + 2) : 65;
{II}if (charCode2 >= base64LookupLen) {{
{III}return new Either<Uint8Array, string>(
{IIII}null,
{IIII}"Expected a valid character from base64-encoded string, " +
{IIIII}`but got at index ${{i + 2}}: ${{text[i + 2]}} (code: ${{charCode2}})`
{III});
{II}}}
{II}const encoded2 = BASE64_LOOKUP[charCode2];
{II}if (encoded2 === 255) {{
{III}return new Either<Uint8Array, string>(
{IIII}null,
{IIII}"Expected a valid character from base64-encoded string, " +
{IIIII}`but got at index ${{i + 2}}: ${{text[i + 2]}} (code: ${{charCode2}})`
{III});
{II}}}

{II}// We map padding to 65, which is the value of "A".
{II}const charCode3 = i + 3 < lenWoPad ? text.charCodeAt(i + 3) : 65;
{II}if (charCode3 >= base64LookupLen) {{
{III}return new Either<Uint8Array, string>(
{IIII}null,
{IIII}"Expected a valid character from base64-encoded string, " +
{IIIII}`but got at index ${{i + 3}}: ${{text[i + 3]}} (code: ${{charCode3}})`
{III});
{II}}}
{II}const encoded3 = BASE64_LOOKUP[charCode3];
{II}if (encoded3 === 255) {{
{III}return new Either<Uint8Array, string>(
{IIII}null,
{IIII}"Expected a valid character from base64-encoded string, " +
{IIIII}`but got at index ${{i + 3}}: ${{text[i + 3]}} (code: ${{charCode3}})`
{III});
{II}}}

{II}bytes[pointer] = (encoded0 << 2) | (encoded1 >> 4);
{II}pointer++;

{II}bytes[pointer] = ((encoded1 & 15) << 4) | (encoded2 >> 2);
{II}pointer++;

{II}bytes[pointer] = ((encoded2 & 3) << 6) | (encoded3 & 63);
{II}pointer++;
{I}}}

// NOTE (mristin, 2022-12-02):
// We expect Uint8Array to silently ignore writes outside of the buffer,
// but we still want to check here in case the underlying platform was flaky about it.
{I}if (bytes.length !== bytesLength) {{
{II}throw new Error(
{III}`Expected bytes to have length ${{bytesLength}}, but got ${{bytes.length}}`
{II});
{I}}}

{I}return new Either<Uint8Array, string>(bytes, null);
}}"""
        ),
        Stripped(
            f"""\
/**
 * Encode a byte array in base64url.
 *
 * This encoding differs from base64 in so far that the characters `+` and `/` are
 * replaced with `-` and `_` and the padding (`=`) is removed, so that the encoded
 * value can be inserted into URIs without escaping.
 *
 * See RFC 4648 for more details:
 * https://www.rfc-editor.org/rfc/rfc4648#section-5
 *
 * @param bytes - to be encoded
 * @returns `bytes` encoded as base64url text
 */
export function base64UrlEncode(bytes: Uint8Array): string {{
{I}return base64Encode(bytes)
{II}.replace(
{III}/[+/=]/g,
{III}m =>
{IIII}m === '+'  ? '-' :
{IIII}m === '/'  ? '_' :
{IIII}'' // The padding `=` is removed.
{II});
}}"""
        ),
        Stripped(
            f"""\
/**
 * Decode a base64url-encoded byte array.
 *
 * This encoding differs from base64 in so far that the characters `+` and `/` are
 * replaced with `-` and `_` and the padding (`=`) is removed, so that the encoded
 * value can be inserted into URIs without escaping.
 *
 * @param text - to be decoded
 * @returns either the array or an error, if `text` is not a valid base64url encoding
 */
export function base64UrlDecode(text: string): Either<Uint8Array, string> {{
{I}let base64Text = text.replace(/[-_]/g, m => m === '-' ? '+' : '/');
{I}const pad = base64Text.length % 4;
{I}if (pad > 0) {{
{II}base64Text += '='.repeat(4 - pad);
{I}}}

{I}return base64Decode(base64Text);
}}"""
        ),
        typescript_common.WARNING,
    ]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()



"""Generate Golang code for de/serialization of enumerations."""

import io
from typing import Tuple, Optional, List

from icontract import ensure

import aas_core_codegen.naming
from aas_core_codegen import intermediate
from aas_core_codegen.common import Error, Stripped, Identifier, indent_but_first_line
from aas_core_codegen.golang import (
    common as golang_common,
    naming as golang_naming,
)
from aas_core_codegen.golang.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


def _generate_model_type_from_string(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the function to translate our internal ``ModelType`` enum to string."""
    # NOTE (mristin, 2023-06-06):
    # Keep in sync with :func:`_generate_enum_from_string`. We copy/pasted and modified
    # the code as creating an enumeration object out of model type turned out to be
    # too tedious and *less* maintainable than copy/pasting. If you keep changing
    # this part of the code base a lot, and have trouble keeping this function in sync
    # with :func:`_generate_enum_from_string`, please consider re-factoring.

    blocks = []  # type: List[Stripped]

    # region From-string-map

    items = []  # type: List[str]
    for cls in symbol_table.concrete_classes:
        literal_name = golang_naming.enum_literal_name(
            Identifier("Model_type"), cls.name
        )

        literal_value = golang_common.string_literal(
            aas_core_codegen.naming.json_model_type(cls.name)
        )

        items.append(f"{literal_value}: aastypes.{literal_name}")

    from_str_map_name = golang_naming.private_constant_name(
        Identifier("model_type_from_string_map")
    )

    items_joined = "\n".join(f"{item}," for item in items)

    name = golang_naming.enum_name(Identifier("Model_type"))

    blocks.append(
        Stripped(
            f"""\
var {from_str_map_name} = map[string]aastypes.{name} {{
{I}{indent_but_first_line(items_joined, I)}
}}"""
        )
    )

    # endregion

    # region From-string-function
    from_str_name = golang_naming.function_name(Identifier("model_type_from_string"))

    blocks.append(
        Stripped(
            f"""\
// Parse `text` as a string representation of [aastypes.{name}].
//
// If not ok, the literal result is undefined.
func {from_str_name}(
{I}text string,
) (literal aastypes.{name}, ok bool) {{
{I}literal, ok = {from_str_map_name}[text]
{I}return
}}"""
        )
    )

    # endregion

    return Stripped("\n\n".join(blocks))


def _generate_model_type_to_string(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the function to translate a string to our internal ``ModelType``."""
    # NOTE (mristin, 2023-06-06):
    # Keep in sync with :func:`_generate_enum_to_string`. We copy/pasted and modified
    # the code as creating an enumeration object out of model type turned out to be
    # too tedious and *less* maintainable than copy/pasting. If you keep changing
    # this part of the code base a lot, and have trouble keeping this function in sync
    # with :func:`_generate_enum_to_string`, please consider re-factoring.

    blocks = []  # type: List[Stripped]

    # region To-string-map

    items = []  # type: List[str]
    for cls in symbol_table.concrete_classes:
        # NOTE (mristin, 2023-03-29):
        # We assume that enumeration literals are specified using ``iota``.

        literal_value = golang_common.string_literal(
            aas_core_codegen.naming.json_model_type(cls.name)
        )

        items.append(literal_value)

    to_str_array_name = golang_naming.private_constant_name(
        Identifier("model_type_to_string_array")
    )

    items_joined = "\n".join(f"{item}," for item in items)

    blocks.append(
        Stripped(
            f"""\
var {to_str_array_name} = [...]string {{
{I}{indent_but_first_line(items_joined, I)}
}}"""
        )
    )

    # endregion

    # region To-string-function

    to_str_name = golang_naming.function_name(Identifier("model_type_to_string"))

    name = golang_naming.enum_name(Identifier("Model_type"))

    blocks.append(
        Stripped(
            f"""\
// Translate `value` from [aastypes.{name}] to a string.
//
// If the value is not valid, the OK is false and the string representation is
// undefined.
func {to_str_name}(
{I}value aastypes.{name},
) (result string, ok bool) {{
{I}i := int(value)
{I}ok =
{II}i >= 0 &&
{II}i < len({to_str_array_name})

{I}if !ok {{
{II}return
{I}}}
{I}result = {to_str_array_name}[value]
{I}return
}}"""
        )
    )

    must_to_str_name = golang_naming.function_name(
        Identifier("must_model_type_to_string")
    )

    blocks.append(
        Stripped(
            f"""\
// Translate the `value` from [aastypes.{name}] to a string.
//
// Panic if the given value is invalid.
func {must_to_str_name}(
{I}value aastypes.{name},
) string {{
{I}result, ok := {to_str_name}(value)
{I}if !ok {{
{II}panic(
{III}fmt.Sprintf(
{IIII}"Invalid value of {name}: %v",
{IIII}value,
{III}),
{II})
{I}}}
{I}return result
}}"""
        )
    )

    # endregion

    return Stripped("\n\n".join(blocks))


def _generate_enum_from_string(enumeration: intermediate.Enumeration) -> Stripped:
    """Generate the functions for de-serializing enumeration from strings."""
    blocks = []  # type: List[Stripped]

    name = golang_naming.enum_name(enumeration.name)

    # region From-string-map

    items = []  # type: List[str]
    for literal in enumeration.literals:
        literal_name = golang_naming.enum_literal_name(enumeration.name, literal.name)
        literal_value = golang_common.string_literal(literal.value)

        items.append(f"{literal_value}: aastypes.{literal_name}")

    from_str_map_name = golang_naming.private_constant_name(
        Identifier(f"{enumeration.name}_from_string_map")
    )

    items_joined = "\n".join(f"{item}," for item in items)

    blocks.append(
        Stripped(
            f"""\
var {from_str_map_name} = map[string]aastypes.{name} {{
{I}{indent_but_first_line(items_joined, I)}
}}"""
        )
    )

    # endregion

    # region From-string-function
    from_str_name = golang_naming.function_name(
        Identifier(f"{enumeration.name}_from_string")
    )

    blocks.append(
        Stripped(
            f"""\
// Parse `text` as a string representation of [aastypes.{name}].
//
// If not ok, the literal result is undefined.
func {from_str_name}(
{I}text string,
) (literal aastypes.{name}, ok bool) {{
{I}literal, ok = {from_str_map_name}[text]
{I}return
}}"""
        )
    )

    # endregion

    return Stripped("\n\n".join(blocks))


def _generate_enum_to_string(enumeration: intermediate.Enumeration) -> Stripped:
    """Generate the functions for serializing enumerations to strings."""
    blocks = []  # type: List[Stripped]

    name = golang_naming.enum_name(enumeration.name)

    # region To-string-map

    items = []  # type: List[str]
    for literal in enumeration.literals:
        # NOTE (mristin, 2023-03-29):
        # We assume that enumeration literals are specified using ``iota``.
        literal_value = golang_common.string_literal(literal.value)

        items.append(literal_value)

    to_str_array_name = golang_naming.private_constant_name(
        Identifier(f"{enumeration.name}_to_string_array")
    )

    items_joined = "\n".join(f"{item}," for item in items)

    blocks.append(
        Stripped(
            f"""\
var {to_str_array_name} = [...]string {{
{I}{indent_but_first_line(items_joined, I)}
}}"""
        )
    )

    # endregion

    # region To-string-function

    to_str_name = golang_naming.function_name(
        Identifier(f"{enumeration.name}_to_string")
    )

    blocks.append(
        Stripped(
            f"""\
// Translate `value` from [aastypes.{name}] to a string.
//
// If the value is not valid, the OK is false and the string representation is
// undefined.
func {to_str_name}(
{I}value aastypes.{name},
) (result string, ok bool) {{
{I}i := int(value)
{I}ok =
{II}i >= 0 &&
{II}i < len({to_str_array_name})

{I}if !ok {{
{II}return
{I}}}
{I}result = {to_str_array_name}[value]
{I}return
}}"""
        )
    )

    must_to_str_name = golang_naming.function_name(
        Identifier(f"must_{enumeration.name}_to_string")
    )

    blocks.append(
        Stripped(
            f"""\
// Translate the `value` from [aastypes.{name}] to a string.
//
// Panic if the given value is invalid.
func {must_to_str_name}(
{I}value aastypes.{name},
) string {{
{I}result, ok := {to_str_name}(value)
{I}if !ok {{
{II}panic(
{III}fmt.Sprintf(
{IIII}"Invalid value of {name}: %v",
{IIII}value,
{III}),
{II})
{I}}}
{I}return result
}}"""
        )
    )

    # endregion

    return Stripped("\n\n".join(blocks))


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable, repo_url: Stripped
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate the Golang code for the de/serialization of strings."""
    aastypes_url_literal = golang_common.string_literal(f"{repo_url}/types")

    blocks = [
        Stripped(
            """\
// Package stringification converts enumerations from and to string representations.
package stringification"""
        ),
        golang_common.WARNING,
        Stripped(
            f"""\
import (
{I}"fmt"
{I}aastypes {aastypes_url_literal}
)"""
        ),
        _generate_model_type_from_string(symbol_table=symbol_table),
        _generate_model_type_to_string(symbol_table=symbol_table),
    ]

    for enum in symbol_table.enumerations:
        blocks.append(_generate_enum_from_string(enumeration=enum))
        blocks.append(_generate_enum_to_string(enumeration=enum))

    blocks.append(golang_common.WARNING)

    out = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            out.write("\n\n")

        out.write(block)

    out.write("\n")

    return out.getvalue(), None
